#[allow(clippy::module_inception)]
pub mod bot {

    use log::{info, debug};
    use matrix_sdk::{
        Client,
        config::{SyncSettings, RequestConfig},
        event_handler::Ctx,
        room::{Joined, Room, RoomMember},
        ruma::{
            UserId, ServerName, RoomId, OwnedEventId,
            events::room::message::{OriginalSyncRoomMessageEvent, RoomMessageEventContent}, 
        }
    };

    use sled;
    use crate::utils::utils::*;
    use crate::config_reader::config_reader::BotUserInfo;

    #[allow(dead_code)]
    #[derive(Clone)]
    pub struct Bot {
        client: Client,
        info: BotUserInfo,
        joined_room: Joined,
        swear_list: Vec<String>,
        database_handle: sled::Db,
        spam_info: SpamInfo
    }

    impl Bot {
        pub async fn bot_login(creds: BotUserInfo) -> Self {
            let user_id = creds.user_id.as_str();
            let password = creds.password.as_str();
            // let homeserver = creds.homeserver.as_str();
            let room = RoomId::parse(creds.room.as_str()).unwrap();
            // let client = Client::builder().homeserver_url(homeserver).build().await.unwrap();

            let user = UserId::parse(user_id).expect("Unable to parse bot user id");
            let server_name = ServerName::parse(user.server_name()).unwrap();
            let request_config = RequestConfig::new().force_auth();
            let client = Client::builder()
                .server_name(&server_name)
                .request_config(request_config)
                .build()
                .await.
                unwrap();

            info!("Logging in…");
            let response = client
                .login_username(user.localpart(), password)
                .initial_device_display_name("Matrix-Modbot")
                .send()
                .await
                .expect("Unable to login");

            info!("Doing the initial sync…");
            client
                .sync_once(SyncSettings::new())
                .await
                .expect("Unable to sync");

            info!(
                "Logged in as {}, got device_id {}",
                response.user_id, response.device_id
            );
            
            // Code snippet from hebbot https://github.com/haecker-felix/hebbot
            // Try to accept room invite, if any
            if let Some(invited_room) = client.get_invited_room(&room) {
                invited_room
                    .accept_invitation()
                    .await
                    .expect("Matrix-Modbot could not join the invited room");
            }

            // Need to do another sync to make sure bot knows the newly joined rooms
            client
                .sync_once(SyncSettings::new())
                .await
                .expect("Unable to sync");

            let joined_room = client
                .get_joined_room(&room)
                .expect("Cannot get joined room");

            let swear_list = create_swear_list(creds.swear_list_url.as_str())
                .expect("Could not download swear list");

            // Syncs members list of the joined room
            joined_room.sync_members().await.unwrap();

            let members_list = client
                .get_joined_room(joined_room.room_id())
                .unwrap()
                .joined_members()
                .await
                .expect("Cannot get members");

            let path = "members_database";
            let db = sled::open(path).unwrap();
            let default_reputation: u8 = 32;
            for member in members_list {
                dbg!(
                    db.insert(member.user_id().as_str(), &default_reputation.to_be_bytes()).unwrap()
                );
            };

            let bot = Bot {
                client,
                info: creds,
                joined_room,
                swear_list,
                database_handle: db,
                spam_info: SpamInfo::new()
            };

            bot.send_message("Sucessfully started Matrix-Modbot").await;
            
            bot.client.add_event_handler_context(bot.clone());
            bot.client.add_event_handler(Self::on_room_message);

            info!("Started syncing");
            bot.client.sync(SyncSettings::new()).await.unwrap();

            bot
        }

        async fn on_room_message(event: OriginalSyncRoomMessageEvent, room: Room, Ctx(bot): Ctx<Bot>) {
            if let Room::Joined(_joined) = &room {
                // Standard text message
                if let Some(text) = get_message_event_text(&event) {
                    let member = room.get_member(&event.sender).await.unwrap().unwrap();
                    let id = &event.event_id;

                    if bot.info.allow_swear && detect_swear_from_message(&bot.swear_list, &text) {
                            bot.delete_message_from_room(id, &member, "Swearing").await;
                        };
                    }
                }

                // Message edit
                if let Some((edited_msg_event_id, text)) = get_edited_message_event_text(&event){
                    if bot.info.allow_swear && detect_swear_from_message(&bot.swear_list, &text) {
                            let member = room.get_member(&event.sender).await.unwrap().unwrap();
                            bot.delete_message_from_room(&edited_msg_event_id, &member, "Swearing").await;
                        };


                    }
        }

        async fn delete_message_from_room(&self, event_id: &OwnedEventId, member: &RoomMember, reason: &str) {
            let result = self.joined_room.redact(event_id, Some(reason), None).await;
            if result.is_err() {
                self.send_message("Error deleting message").await;
            };
            self.send_message(&format!("{} is not permitted on this server", reason)).await;
            self.subtract_reputation_from_member(member).await.expect("Failed to subtract reputation from member");
        }

        async fn subtract_reputation_from_member(&self, member: &RoomMember) -> sled::Result<()> {
            
            let member_id: &str = member.user_id().as_str();
            let mut current_reputation = dbg!(self.database_handle.get(member_id).unwrap().unwrap().first().unwrap().to_owned());
            if current_reputation == 0 {
                self.kick_user(member, "Reputation too low!").await;
                return Ok(())
            }
            else {
                current_reputation -= 1;
                dbg!(self.database_handle.insert(member_id, &current_reputation.to_be_bytes()).unwrap());
                self.send_message(&format!("{}'s current reputation is {}", member_id, current_reputation)).await;
                return Ok(())
            }
        }

        async fn kick_user(&self, member: &RoomMember, reason: &str) {
            let member_id: &str = member.user_id().as_str();
             if member.power_level() <= 50 { // Won't kick mods and admins
                if let Some(_) = self.joined_room.kick_user(member.user_id(), Some(reason)).await.ok() {
                    dbg!(self.database_handle.remove(member_id).unwrap());
                    self.send_message(&format!("Member {} has been kicked.", member_id)).await;
                };
            }
            else {
                self.send_message("Cannot kick moderators and admins").await;
            }
        }
            
        async fn send_message(&self, message: &str) {
            debug!("Send message ({:?}): {}", &self.joined_room, message);
            let content = RoomMessageEventContent::text_plain(message);
            let _ = &self.joined_room.send(content, None)
                .await
                .expect("Unable to send message");
        }

        async fn detect_spam(&self, event_id: &OwnedEventId) -> bool {
            todo!()
        }

    }
}
